/*
	detuned.h
	Copyright Risto Holopainen 2021
	
	help functions, classes for the generator of Auto-detuned
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

*/

#ifndef DETUNED_H
#define DETUNED_H


#include "wipos.h"
#include <deque>
#include "../util/biquad2.h"



using std::deque;

namespace detune
{



class unwiggler
{
	public:
		unwiggler(int sr=48000);
		~unwiggler();
		
		double unwiggle(deque<double> &u);
		void set_g(double GG);
		void set_a(double FF);
		
	private:
		int sr;
		LP3<double> LP;
		int N;
		int n;
		double g;	// delay gain (-1,1)
		allpass1p *alp;
		bool warn;
		double outarange_a, outarange_g;
};


unwiggler::unwiggler(int fs)
{
	sr = fs;
	n=0; 
	g=-0.95; 
	alp = new allpass1p(sr, 1200);
	warn = false;
}

unwiggler::~unwiggler()
{ 
	alp=0; 
	delete alp; 
	if(warn)
		printf("unwiggler: warning, g = %.3f, a = %.3f\n\n",
			outarange_g, outarange_a);
}

void unwiggler::set_g(double GG) 
{ 
	if(fabs(GG) >= 1)
		{
		warn = true;
		outarange_g = GG;
		printf("unwiggler warning: set_g = %.3f\n", GG);
		}
	if(GG > 0)
		g = .5 * (GG + sqrt(GG));
	else
		g = GG;
}

void unwiggler::set_a(double FF)
{
	if(FF <= 0 || FF >= sr/2)
		{
		warn = true;
		outarange_a = FF;
		printf("unwiggler warning: set_a = %.3f\n", FF);
		}
		
	alp->cfr(FF); 
}


double unwiggler::unwiggle(deque<double> &u)
{
	// read u in circle, erase each sample when read
	// rather: mult by g, |g|<1 for decay like karplus+strong

	N = u.size();
	if(N == 0)
		return 0;

	if(++n >= N)
		n = 0;

	double x = u[n];
	u[n] = g * LP.filter(alp->filter(u[n]));
	return x;
}



struct parameters
{
	unsigned int p1;
	unsigned int p2;
};




class detuned
{
	public:
		enum TIMBRE {PINK, SAW, PRIME, TRI, PULSE};
		detuned(int sr=48000, double c0=20, TIMBRE TT=TRI);
		~detuned();
		double gen2();
		void set_periods(const parameters &p);
		void set_frq(double f);
		void set_quantize(unsigned int steps);
		
	private:
		int sr;
		double frq1, frq2;
		double th1, th2, delta;
		double phi_1, phi_2;
		double sn1, sn2;
		double omega, W;
		double Q2(double); // new quantize
		float Q_bias[4];
		unsigned int nq, Qsteps;
		double inv_Qsteps;
		double c;
		double F, sqF;		// new parameters
		unsigned int G, H;
		LP3<double> LP;
		wipos *os_a, *os_b;
};

detuned::detuned(int fs, double c0, TIMBRE TT)
{
	sr = fs;
	th1 = th2 = phi_1 = phi_2 = sn1 = sn2 = 0;
	delta = 0;
	W = 2*M_PI/sr;
	omega = 440*W;
	F = 440;
	G = H = 2;
	c = c0;
	sqF = 1;
	
	switch (TT)
	{
		case TRI :
		{ os_a = new wipos(sr, wipos::TRI);	os_b = new wipos(sr, wipos::TRI); break; }
		case SAW :
		{ os_a = new wipos(sr, wipos::SAW); os_b = new wipos(sr, wipos::SAW); break; }
		case PRIME :
		{ os_a = new wipos(sr, wipos::PRIME); os_b = new wipos(sr, wipos::PRIME); break; }
		case PINK :
		{ os_a = new wipos(sr, wipos::PINK); os_b = new wipos(sr, wipos::PINK); break; }
		case PULSE :
		{ os_a = new wipos(sr, wipos::PULSE); os_b = new wipos(sr, wipos::PULSE); break; }
		default :
		{ os_a = new wipos(sr, wipos::TRI); os_b = new wipos(sr, wipos::TRI); }
	}

// bias index and table
	nq = 0;
	Q_bias[0] = 1.5;
	Q_bias[1] = 1.0; 
	Q_bias[2] = 0.5;
	Q_bias[3] = 2.0;
	Qsteps = 2;
	inv_Qsteps = 1/2.;
}


detuned::~detuned()
{
	delete os_a;
	delete os_b;
}


double detuned::Q2(double x)
{
	return (int(Qsteps*x) * inv_Qsteps);
}


void detuned::set_quantize(unsigned int steps)
{
	Qsteps = steps;
	inv_Qsteps = 1.0 / Qsteps;
}


double detuned::gen2()
{
	frq1 = omega + delta;
	frq2 = omega - delta;
	th1 += W * frq1;
	th2 += W * frq2;
	phi_1 = th1/(G + sqF);
	phi_2 = th2/(H + sqF);
	sn1 = sin(phi_1);
	sn2 = sin(phi_2);
	
	omega = F*Q2(fabs(sn1) + sn2 + Q_bias[nq]);
	omega = LP.filter(omega);
	delta += c*fabs(sn1 + sn2) - delta;
	double x = os_a->gen(frq1) + os_b->gen(frq2);
	return .4*x;
}


void detuned::set_frq(double f)
{
	F = f;
	if(f < 0)
		{
		printf("detuned::set_frq NEGATIVE: %.3f\n", f);
		F = -F;
		}
	sqF = sqrt(F);

	if(++nq >= 4)	// quantization bias
		nq = 0;	
}


void detuned::set_periods(const parameters &u)
{
	// p > 1
	int a = (u.p1 > 1) ? u.p1 : 1;
	int b = (u.p2 > 1) ? u.p2 : 1;
	G = 1 + a;
	H = 1 + b;
}


void update_p2(parameters &Z, const double x)
{
	// suppose x in [0,1]
	// empirical: x min-max 0.000--0.256,	 avg 0.0348
	double u = 2*x-1;
	if(u < -1) u = -1;
	if(u >  1) u = +1;

	double pp = exp(-5*x);
	double qq = sqrt(x + 0.00001);

	double g = 1500*(pp + .75*qq);
	double h =  750*(qq + u*u) + 300;
	Z.p1 = int(g);
	Z.p2 = int(h);	
}


double update_f2(const double x)
{
	// new frequency from x [0,1] -- typically 0 < x < 0.3
	// int q = 25*round(200*x) - 9*exp(6*x) - round(100*x)
	int q = 23*int(225.0*x) + 9*(exp(6.0*x) - int(100.0*x));
	return 20 + q;
}





}	// end namespace


#endif
