/*
	lineclox.cpp
	Copyright Risto Holopainen 2021

	generator of the composition
	speed of thought
	
	WARNING:
	this program is VERY slow, due to the use of 
	a collection of experimental and unreliable random number generators

	outputs three soundfiles intended to be mixed 
	or at least played simultaneously

	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

*/

#include "variclox.h"
#include "line.h"
#include "../util/delay_t.h"
#include <cstdio>
#include <sndfile.h>


void update_param(double &slope, double &lo, double &hi, double range=0.25)
{
	slope = range*trand();
	lo = .4*(rund()  - 1);
	hi = .5*(trind() + 1);
}



class synt
{
	public:
		synt(int L);
		void update_lim(int T, double r=.25) {updatelim = T; range=r;}
		~synt();
		void gen(double *buf, double *nos);
		
	private:
		double sl, lo, hi;
		int t;
		int L;
		double *x, u;
		line ell;
		double G, y;
		int upcount;
		int updatelim;
		double range;
};


synt::synt(int sz)
{
	L = sz;
	x = new double[L];
	u = 0.0;
	sl = 0.0;
	lo = -1;
	hi = +1;
	t = 0;
	ell.set(sl, lo, hi);
	upcount = 0;
	updatelim = 15000;	// how often to change params on average
	range = 0.25;
}


synt::~synt()
{
	x = 0;
	delete x;
	printf("updates: %d \n", upcount);
}


void synt::gen(double *buf, double *nos)
{

	if( (t += toss_m(10)) > updatelim)
		{
		update_param(sl, lo, hi, range);
		t = 0;
		ell.set(sl, lo, hi);
		++upcount;
		}

	ell.draw(x, L);

	if(toss())
		{
		int m = 1 + toss_m(9);
		if(m > 6 && lo > -0.9 && hi < 0.8)
			{
			lo *= 1.07;
			hi *= 1.11;
			sl *= 0.97;
			ell.set(sl, lo, hi);
			}
		G = trix();
		if(t % 64000 == 1)
			printf("G = %+.3f \n", G);
		for(int i=0; i<L; i++)
			{
			y = gaussian(m);
			nos[i] = y;
			x[i] += G * y;
			if(nos[i] > +0.9) nos[i] = +0.9;
			if(nos[i] < -0.9) nos[i] = -0.9;
			}
		}
	else
		for(int i=0; i<L; i++)
			nos[i] *= 1/5.;

	// integrate
	for(int i=0; i<L; i++)
		{
		u += .125*x[i];
		u *= 0.9995;
		if (u > 1) u = 1;
		else if(u < -1) u = -1;
		buf[i] = .75*u + .05*x[i];
		}
}



class echo
{
	public:
		echo(int bufsz, int delsz);
		~echo();
		double process(const double *x, double *y);	// return amp of x
		
	private:
		int L;
		double u, v, A, s;
		delay <double> *del;
};

echo::echo(int bsz, int dsz)
{
	L = bsz;
	u = v = 0.0;
	del = new delay<double>(dsz, 0.0);
	A = 0;
}

echo::~echo()
{
	del = 0;
	delete del;
}


double echo::process(const double *x, double *y)
{
	s = 0;
	for(int i=0; i<L; i++)
		{
		s += x[i]*x[i];
		u = .8*x[i] + .2*u;
		v = .6*u + .4*del->z(v) + .2*v;
		y[i] = v;
		}
	return s / L;
}


void bufmult(const double *A, double *B, int sz)
{
	for(int i=0; i<sz; i++)
		B[i] *= A[i];
}

void bufmult(const double *a, const double *b, double *c, int sz)
{
	for(int i=0; i<sz; i++)
		c[i] = a[i] * b[i];
}

void mux(const double *u, const double *v, int sz, double *st)
{
	// u,v -> stereo
	for(int i=0, k=0; i<sz; i++, k += 2)
		{
		st[k  ] = u[i];
		st[k+1] = v[i];
		}
}


class envelope
{
	public:
		envelope(int delaytime);
		~envelope();
		void env(double *x, int sz);
		void slope(double sl);
		
	private:
		delay<double> *dd;
		double A, g, amp;
		line *lix;
		double slp, min, max;
};

envelope::envelope(int ds)
{
	dd = new delay<double>(ds, 0.0);
	min = -0.05;
	max = 0.95;
	slp = 1/64000.;
	lix = new line();
	lix->set(slp, min, max);
	A = g = amp = 0.0;
}

envelope::~envelope()
{
	dd = 0;
	delete dd;
	lix = 0;
	delete lix;
}

void envelope::env(double *x, int sz)
{
	// modulate, echo, mult
	
	for(int i=0; i<sz; i++)
		{
		A = lix->draw();
		g = dd->z( .8*A - .2*g );
		amp = .6*A + .4*g;
		x[i] *= amp;
		}
}

void envelope::slope(double sl)
{
	lix->set(sl, min, max);
}

// see, we don't do software engineering here, this one is global
int trigcount = 0;
bool trig(double a)
{
	++trigcount;
	// not too often
	double psi = trix();
	return (psi > 0 && a > psi);
}

int upmods = 0;		// global for stats
void update_mods(double p[3])
{
	double sl = p[0], lo = p[1], hi = p[2];

	if(hi - lo > 0.5)
		{
		lo += 0.03;
		hi -= 0.02;	
		}
	else
		{
		lo = -0.1;
		hi = 1.0;	
		}

	sl *= 0.9;
	sl += (trand() - .5) / 75000;
	if(sl < 1/125000.)
		{
		sl = trind() / 32000.;
		}
	p[0] = sl;
	p[1] = lo;
	p[2] = hi;
	++upmods;
}

int sections(int k, int sect[3])
{
	// cumulative time
	int m = 128;
	int tot = 0;
	int add = 0;
	int pm = 32;

	for(int i=0; i<3; i++)
		{
		add = toss_m(pm << i);
		sect[i] = k*(m + add);
		tot += sect[i];
		printf("%d add %d, m %d\n", i, add, m);
		m *= 3;
		m /= 4;
		sect[i] += sect[i-1];
		}

	return tot;
}

int main()
{
	time_t begins_now = time(NULL);
	const int sr = 48000;	
	line slowline, modline;
	double slowparams[3] = {.7/sr, -.93, .98};	
	double moduparams[3] = {.3/sr, -.05, 1.0};	
	slowline.set(slowparams);
	modline.set(moduparams);

	const int N = 24;
	double af[N], bf[N], mf[N], cn[N], dn[N], ns[N], mn[N];
	double st[2*N];
	double env[N], amp[N], lfo[N];
	double A=0;
	synt SY(N), SX(N), SZ(N);
	int updates_z = 75000;
	double range_z = .025;
	SZ.update_lim(updates_z, range_z);	// slow & small
	echo EK(N, sr/3);
	envelope velopen(2.13*sr);

	SNDFILE *ouf, *nof, *caf;
	SF_INFO head1, head2;
	head1.samplerate = sr;
	head1.channels = 2;
	head1.format = (SF_FORMAT_WAV | SF_FORMAT_PCM_24);
	head2 = head1;
	ouf = sf_open("linecloxer.wav", SFM_WRITE, &head1);
	nof = sf_open("linecloxnoise.wav", SFM_WRITE, &head2);
	caf = sf_open("lineclox_center.wav", SFM_WRITE, &head1);

	int sect[3];
	const int TL = sections(sr/N, sect);
	printf("TL : %3d sec, sections: %3d s, %3d s, %3d s \n",
		TL*N/sr, sect[0]*N/sr, sect[1]*N/sr, sect[2]*N/sr);


	for(int i=0, k=0; i<TL; i++)
		{
		if (i == sect[k])
			{
			++k;
			printf("begin section %d \n", k);
			if(k==1)
				{
				range_z = 0.75;
				updates_z = 900;	// was outside if
				SZ.update_lim(updates_z, range_z);	// ditto
				}
			}
		if(k == 2 && trig(A))	// slow down SZ, speed up rest
			{
			if(updates_z < 175000)
				++updates_z;
			if(range_z > 0.05)
				range_z *= 0.99998;
			SZ.update_lim(updates_z, range_z);
			}
			
		
		slowline.draw(env, N);
		modline.draw(amp, N);
		SY.gen(bf, cn);
		SX.gen(af, dn);
		
		SZ.gen(mf, mn);
		
		bufmult(env, bf, N);
		bufmult(amp, af, N);
		bufmult(env, amp, lfo, N);	// lfo: center, sz
		bufmult(lfo, mn, N);
		mux(mf, mn, N, st);
		sf_write_double(caf, st, 2*N);
		

		mux(af, bf, N, st);
		sf_write_double(ouf, st, 2*N);
		
		if(trig(A))		// retune lines
			{
			update_mods(slowparams);
			slowline.set(slowparams);
			double nextslope = rund()/sr;
			velopen.slope(nextslope);
			}
		
		A = EK.process(cn, ns);
		if(trig(A))
			{
			update_mods(moduparams);
			modline.set(moduparams);
			}
		
		bufmult(amp, ns, N);
		velopen.env(dn, N);
		mux(ns, dn, N, st);
		sf_write_double(nof, st, 2*N);
		}
	
	sf_close(ouf);
	sf_close(nof);
	sf_close(caf);
	
	printf("%d trigs, \t%d upmods, \t %d update_z\n", trigcount, upmods, updates_z);

	time_t ends_now = time(NULL);
	double workday = difftime(ends_now, begins_now);

	printf("start: \t%ld \nend:  \t%ld \nworking hard for %.0f seconds\n\n",
		begins_now, ends_now, workday);
		

	return 0;
}
