/*
	timbre.cpp
	Copyright Risto Holopainen 2021
	
	Generator of the composition The Wastebasket of Timbre
	outputs four soundfiles intended to be mixed and played simultaneously
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

*/

#include "timbre.h"
#include "../util/envelope.h"
#include <sndfile.h>


using namespace waveform;



void q_delay(const double f[4], double g[4])
{
	for(int i=0; i<4; i++)
		g[i] = (f[i] < 1200) ? f[i] : f[i] / 4;
}


void mix(const double x[4], const double y[4], double st[2])
{
	double L=0, R=0;
	for(int i=0, j=1; i<4; i += 2, j += 2)
		{
		L += x[i] + y[i];
		R += x[j] + y[j];
		}
	st[0] = .25*L;
	st[1] = .25*R;
}

void quadadd(double F[4], double v)
{
	for(int i=0; i<4; F[i] += v, ++i) ;
}

void vectmult(const double A, double *x, int sz)
{
	for(int i=0; i<sz; x[i] *= A, ++i) ;
}


void xfade(const double a, double *s)
{
	s[0] *= a;
	s[1] *= (1-a);
	s[2] *= -a;
	s[3] *= (a-1);
}




class quadenvelope
{
	public:
		quadenvelope(int sr, const double c[4]);
		~quadenvelope();
		void mult(double q[4]);
		void duties(const double D[4]);
		double get_duty() const {return duty;}
		void cycles(const double C[4]);
		void get_cycles(double cc[4]) const ;
	
	private:
		int sr;
		double duty;
		deque<envelope> envie;
};

quadenvelope::quadenvelope(int sr, const double cc[4])
{
	duty = 0.36;
	envie.push_back(envelope(sr, SQUARE, cc[0], duty));
	envie.push_back(envelope(sr, SQUARE, cc[1], duty));
	envie.push_back(envelope(sr, EXPO, cc[2], duty));
	envie.push_back(envelope(sr, RAMP, cc[3], duty));				
}

quadenvelope::~quadenvelope() {;}


void quadenvelope::mult(double q[4])
{
	for(int i=0; i<4; i++)
		q[i] *= envie[i].env();
}


void quadenvelope::duties(const double D[4])
{
	for(int i=0; i<4; i++)
		envie[i].duty(D[i]);
}

void quadenvelope::cycles(const double C[4])
{
	for(int i=0; i<4; i++)
		envie[i].cycle(C[i]);	
}

void quadenvelope::get_cycles(double cc[4]) const 
{
	for(int i=0; i<4; i++)
		cc[i] = envie[i].get_cycle();
}

class percussion
{
	public:
		percussion(int sr);
		~percussion();
		double play(const double x[4]);
		
	private:
		int sr;
		envelope *dec1, *dec2;
		lopass *lp1, *lp2;
		double u, v, A1, A2;
		double c1, c2;
		int m, n;
};


percussion::percussion(int fs)
{
	sr = fs;
	c1 = 5/7.;
	c2 = 2/7.;
	dec1 = new envelope (sr, RAMP, c1, 0.4);
	dec2 = new envelope (sr, EXPO, c2, 0.7);
	lp1 = new lopass(sr, 2500);
	lp2 = new lopass(sr, 2500);	
	m = 0; n = 1;
	A1 = 0; A2 = 0;
}

percussion::~percussion()
{
	dec1 = dec2 = 0;
	lp1 = lp2 = 0;
	delete dec1; delete dec2; delete lp1; delete lp2;
}

double percussion::play(const double x[4])
{
	A1 = 1.0-dec1->env();
	A2 = dec2->env();
	u = A1*x[0];
	v = A2*x[1];
	u = lp1->filter(u);
	v = lp2->filter(v);
	
	if(++m > 33*sr*(1+A1))
		{
		c1 += 1/7.;
		dec1->cycle(c1);
		if(c2 > 9/7.)
			c2 = 1/7.;
		m = 0;
		}
	if(++n > 36*sr*(1+A2))
		{
		n = 0;
		c2 += 1/7.;
		dec2->cycle(c2);
		if(c1 > 4/7.)
			c1 = 1/7.;
		}
	
	return u*v;
}



void duty_handler(quadenvelope &eg, float w)
{
	double dt = eg.get_duty();

	if(dt > 0.01 && w < 0.7)
		dt *= 0.975;
	if(w > 0.8 && dt < 0.4)
		dt += 0.025;

	double hh[4] = {dt, dt, .75*dt, .75*dt};
	eg.duties(hh);
}



void cycle_stretcher(quadenvelope &eg, float w)
{
	double cc[4], eps = 1/18.;
	eg.get_cycles(cc);
	if(w>0.5)
		{
		for(int i=0; i<4; cc[i] += eps, ++i)
			if(cc[i] > 9)
				cc[i] = 1/15.;
		}
	eg.cycles(cc);
}


void check_pitch(double F[4])
{
	for(int i=0; i<4; i++)
		if(F[i] > 1400)
			F[i] /= 4;
}


void grand_mixer(const double a[4], const double b[4], const double c[4], 
				const double d[4], const double e[4],  double mono, 
				const double st[2], const double su[2], double *out)
{
	double L = mono + st[0] + su[0];
	double R = mono + st[1] + su[1];

	for(int i=0, j=1; i<4; i += 2, j += 2)
		{
		L += a[i] + b[i] + c[i] + d[i] + e[i];
		R += a[j] + b[j] + c[j] + d[j] + e[j];
		}
	double norm = 1/15.;
	
	out[0] = norm*L;
	out[1] = norm*R;
}


void home_mixer(const double a[4], const double b[4], double *out)
{
	double L = 0;
	double R = 0;
	for(int i=0, j=1; i<4; i += 2, j += 2)
		{
		L += a[i] + b[i];
		R += a[j] + b[j];
		}
	
	out[0] = .2*L;
	out[1] = .2*R;
}


void other_mixer(const double q[4], const double sn[2], double *outs)
{
	//Q5, xsin, st,
	double L = q[0] + q[2] + sn[0];
	double R = q[1] + q[3] + sn[1];
	
	outs[0] = .25*L;
	outs[1] = .25*R;
}

int main()
{
	pitchline pp, qq;
	make_seq_G(pp, 11, 2);
	make_seq_Ciss(qq, 11, 3);
	sequencer<double> seq(qq), sep(pp);
	const int sr = 48000;
	
	SNDFILE *ouf12, *ouf34, *other, *monop;
	SF_INFO head, mhead;
	head.samplerate = sr;
	head.channels = 2;
	head.format = (SF_FORMAT_WAV | SF_FORMAT_PCM_24);
	mhead = head; mhead.channels = 1;
	ouf12 = sf_open("timbral_Q12.wav", SFM_WRITE, &head);
	ouf34 = sf_open("timbral_Q34.wav", SFM_WRITE, &head);
	other = sf_open("timbral_other.wav", SFM_WRITE, &head);
	monop = sf_open("timbral_perk.wav", SFM_WRITE, &mhead);

	const int L = 4097;
	double wtb[L], w0[L], w1[L], w2[L], w3[L];
	make_sine(wtb, L);
	make_triangle(w0, L, 31);
	make_sawtooth(w1, L, 29);
	make_order3(w2, L, 31);
	make_pulse(w3, L, -0.1, 31);
	oscil Osc1(sr, wtb, L);
	oscil Osc2(sr, wtb, L);

	double xsin[2]={0,0}, sinefrq[2]={1,1}, sfr[2]={1,1};
	double F1[4]={0,0,1,1}, Q1[4]={0};
	double F2[4]={1,2,3,4}, Q2[4]={0};	
	double F3[4]={1,2,3,4}, Q3[4]={0};			// nasal
	double F4[4]={1,2,3,4}, Q4[4]={0}; 			// forma
	double F5[4]={1,2,3,4}, Q5[4]={0};
	orchestra quart(sr, w0, w1, w2, w3, L);


	make_prime(w0, L, 11);
	make_sparse(w1, L, 55);
	make_symbump(w2, L);
	make_pink(w3, L, 29);
	orchestra quint(sr, w0, w1, w2, w3, L);
	
	make_halfsine(w0, L, 31);
	make_pulse(w1, L, 0.85, 31);
	make_diffbump(w2, L);
	make_volterra(w3, L);
	orchestra nasal(sr, w0, w1, w2, w3, L);		// wants filter
	
	make_formant(w0, L, 31, 7);
	make_twinpeaks(w1, L);
	make_band(w2, L, 3, 11);
	make_comb(w3, L, 4, 31);
	orchestra forma(sr, w0, w1, w2, w3, L);
	
	make_darboux(w0, L);
	make_octaves(w1, L);
	make_expogliss(w2, L, 7);
	make_diphone(w3, L, 6);
	orchestra ufono(sr, w0, w1, w2, w3, L);

	envelope eg(sr, RAMP, 2.75, 0.3);
	envelope perkenv(sr, SQUARE, 17.5, 0.5);
	envelope fiven(sr, RAMP, 7, 1/3.);
	
	double periods_1[4] = {6/5., 7/5., 8/5., 9/5.};
	double periods_2[4] = {12/10., 12/9., 12/8, 12/7.};
	double periods_3[4] = {3/2., 5/3., 7/3., 11/5.};
	double periods_4[4] = {7/4., 2, 9/4., 5/2.};
	quadenvelope vecu1(sr, periods_1), vecu2(sr, periods_2);
	quadenvelope vecu3(sr, periods_3), vecu4(sr, periods_4);
	percussion perk(sr);

	double vibmod, amod, perksmp, amp;
	quark qwerty_1(1), qwerty_2(2), qwerty_4(4);	
	const int K = 1 << 7;
	double buf[2*K] = {0};
	double puf[2*K], bof[2*K], monobuf[K];
	const int kr = sr/K;
	const int N = 475*kr;

	vibrato LFO(kr), AMLO(kr);
	
	int n=0;
	while(n++ < N)
		{
		float w = (float) n/N;
		if(n % kr/3 == 0)
			{
			sfr[0] = sep.tick();
			sfr[1] = seq.tick();
			qwerty_1.step(F1);
			qwerty_2.step(F3);
			cycle_stretcher(vecu1, w);
			cycle_stretcher(vecu2, w);
			cycle_stretcher(vecu3, w);
			duty_handler(vecu1, 1.2*w);	// tricking it.
			}
		if(n % (7*kr/5) == 0)
			{
			q_delay(F1, F2);
			qwerty_4.step(F4);
			q_delay(F3, F5);
			check_pitch(F4);
			check_pitch(F3);
			if(n > N/5)
				{
				check_pitch(F1);
				duty_handler(vecu2, w);
				duty_handler(vecu3, w);
				duty_handler(vecu4, w);
				}
			}
		portamento(sfr, sinefrq, w);
		vibmod = LFO.vib(w);
		quadadd(F1, vibmod);
		amod = AMLO.vib(1-w);
		
		for(int i=0, k=0; i<K; i++, k += 2)
			{
			amp = eg.env();	
			xsin[0] = amp*Osc1.stable(sinefrq[0]);
			xsin[1] = amp*Osc2.stable(sinefrq[1]);
			vectmult(amp, xsin, 2);
			
			quart.toot(F1, Q1);
			quint.toot(F2, Q2);
			nasal.toot(F3, Q3);	// to filter
			forma.toot(F4, Q4);
			ufono.toot(F5, Q5);
			
			perksmp = perk.play(Q2);
			perksmp *= perkenv.env();
			vecu1.mult(Q1);
			vecu2.mult(Q2);
			vecu3.mult(Q3);
			vecu4.mult(Q4);

			xfade(amod, Q5);
			vectmult(fiven.env(), Q5, 4);
						
			/*
			grand_mixer(Q1, Q2, Q3, Q4, Q5, perksmp, xsin, st, outs);
			*/
			
			home_mixer(Q1, Q2, buf+k);
			home_mixer(Q3, Q4, puf+k);
			other_mixer(Q5, xsin, bof+k);
			monobuf[i] = perksmp;
			}
		sf_write_double(ouf12, buf, 2*K);
		sf_write_double(ouf34, puf, 2*K);
		sf_write_double(other, bof, 2*K);
		sf_write_double(monop, monobuf, K);
		}

	sf_close(ouf12);
	sf_close(ouf34);
	sf_close(other);
	sf_close(monop);
	return 0;
}


