/*
	envelope
	Copyright Risto Holopainen 2021
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "../util/boxcar.h"

enum shapes{SQUARE, RAMP, EXPO};



class envelope
{
	public:
		envelope(int sr);
		envelope(int sr, shapes sh, double cyc, double dutt);
		~envelope();
		void shape(shapes SH);
		void cycle(double seconds);
		void duty(double h);	// 0--1
		double get_duty() const {return Duty;}
		double get_cycle() const {return (double)Cycle/sr;}
		double env();

		// internal use
		typedef void (envelope::*fp)();	// points to below methods
		void square();
		void ramp();
		void expo();
		
	private:
		fp how;
		int sr;
		uint n;
		uint t0;	// on duration
		uint Cycle;
		double Duty;
		double A, g, lambda;
		boxcar *BX;
};


envelope::envelope(int fs)
{
	sr = fs;
	n = 0;
	t0 = .1*sr;
	Cycle = 2*t0;
	Duty = t0 / Cycle;
	BX = new boxcar(25);
	g = 1.0 / t0; 	// ramp 0->1 in t0 smps
	lambda = exp(-log(8.) / t0);
	how = &envelope::square;
}

envelope::envelope(int fs, shapes sh, double cyc, double dt)
{
	sr = fs;
	n = 0;
	t0 = dt * cyc;
	Duty = (dt > 0 && dt < 1) ? dt : 0.5;	
	cycle(cyc);
	duty(dt);

	BX = new boxcar(25);
	g = 1.0 / t0; 	// ramp 0->1 in t0 smps
	lambda = exp(-log(8.) / t0);
	shape(sh);
}


envelope::~envelope()
{
	BX = 0;
	delete BX;
}


double envelope::env()
{	
	(this->*how)();
	if(++n > Cycle)
		n = 0;

	return BX->filter(A);
}


void envelope::square()
{
	A = (n < t0) ? 1 : 0;
}


void envelope::ramp()
{
	A = (n < t0) ? g + A : 0;
}


void envelope::expo()
{
	if(n < 15)
		{
		A = 1.0;
		}
	else
		{
		A *= lambda;
		}
}

void envelope::shape(shapes shp)
{
	switch (shp)
		{
		case SQUARE :
			{
			how = &envelope::square;
			break;
			}
		case RAMP :
			{
			how = &envelope::ramp;
			break;
			}
		case EXPO :
			{
			how = &envelope::expo;
			break;
			}
		}
}

void envelope::cycle(double sec)
{
	int cc = sr*sec;
	if(cc > 0)
		{
		Cycle = cc;
		t0 = Duty*cc;
		g = 1.0 / t0;
		lambda = exp(-log(8) / t0);
		}
}

void envelope::duty(double h)
{
	double d = (h > 0 && h < 1) ? h : 0.5;
	Duty = d;
	t0 = d*Cycle;
	g = 1.0 / t0; 				// ramp 0->1 in t0 smps
	lambda = exp(-log(8) / t0);	// decay to 1/8 in t0
}

